/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/input/joystick.h"
using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

#include "../../utils/misc/framerate.h"


// Misc utility

float	flerp(float t, float f0, float f1)
{
	return (1 - t) * f0 + t * f1;
}


float	fclamp(float val, float min_val, float max_val)
{
	if (val < min_val) return min_val;
	if (val > max_val) return max_val;
	return val;
}

float	fmax(float a, float b)
{
	if (a > b) return a;
	return b;
}


void	flowpass(float* in_out, float c0, float new_val)
{
	*in_out = *in_out * c0 + new_val * (1 - c0);
}


/********************************* globals **********************************/

// window/camera stuff
char const AppName[] = "IGJ2 - Deadly Dance Of The Robots";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 50.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
Sim simulator;
Drag_Force    global_drag(.05f, .05f);    //a little drag helps with stability
Gravity_Force gravity(Vec2(0.0f, -1.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);

char *TestLevelFileName = "bodies.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    float Gravity;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;

	float	m_filter_damping;
	float	m_balance_gain;
	float	m_balance_offset;

	float	x0;
	float	x1;
	float y0;
	float y1;

	int	sound_index;

	global_properties()
	{
		Gravity = -100.0f;

		MouseSpringKs = 128.0f;
		MouseSpringKd = 32.0f;
		MouseSpringRestLength = 0.0f;

		m_filter_damping = 0.99f;
		m_balance_gain = 5000.0f;
		m_balance_offset = 10.0f;

		x0 = -50;
		y0 = -40;
		x1 = 50;
		y1 = 40;

		sound_index = 0;
	}
};
static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

static bool MovingCamera = false;

// game stuff
struct game_object
{
    game_object( char *texture_file , Rigid_Body *rbInit ) :
            rb(rbInit)
	{
		assert(rb);

		simulator.add(*rb);
	}

    ~game_object( void )
    {
		delete rb;
	}
    
    Rigid_Body *rb;
};


//
// State.
//
static vector<game_object*> objs;
static vector<Binary_Force *> forces;

static Vec2 window_mouse, world_mouse;
static Joystick PS2Controller;

static Binary_Force *spring = 0;

static float	s_leg_control[2] = { 0, 0 };	// left/right input
static bool	s_up_key = false;
static bool	s_down_key = false;


// static bool	s_left_u_key = false;
// static bool	s_left_d_key = false;
// static bool	s_left_lu_key = false;
// static bool	s_left_ld_key = false;
// static bool	s_left_ru_key = false;
// static bool	s_left_rd_key = false;
// static bool s_left_punch_key = false;

// static bool	s_right_lu_key = false;
// static bool	s_right_ld_key = false;
// static bool	s_right_ru_key = false;
// static bool	s_right_rd_key = false;
// static bool s_right_punch_key = false;


enum key_id
{
	K_U,
	K_D,
	K_LU,
	K_LD,
	K_RU,
	K_RD,
	K_PUNCH,
	K_SWING,

	KEY_COUNT
};

static bool	s_any_start_key(false);
static bool	s_left_keys[KEY_COUNT], s_right_keys[KEY_COUNT];


enum pose_dofs
{
	LEFT_HIP_ANGLE = 0,
	LEFT_HIP_KS,
	LEFT_KNEE_ANGLE,
	LEFT_KNEE_KS,

	RIGHT_HIP_ANGLE,
	RIGHT_HIP_KS,
	RIGHT_KNEE_ANGLE,
	RIGHT_KNEE_KS,
		
	DOF_COUNT
};

	
struct pose_data
{
	// Floats only in here, one per DOF_COUNT!
	float	m_left_hip_angle;
	float	m_left_hip_ks;
	float	m_left_knee_angle;
	float	m_left_knee_ks;
	
	float	m_right_hip_angle;
	float	m_right_hip_ks;
	float	m_right_knee_angle;	
	float	m_right_knee_ks;
};


void	pose_data_madd(pose_data* accum, const pose_data& a_in, float weight)
{
	assert(sizeof(pose_data) == sizeof(float) * DOF_COUNT);

	float*	result = (float*) accum;
	const float*	a = (const float*) &a_in;

	for (int i = 0; i < DOF_COUNT; i++)
	{
		result[i] += a[i] * weight;
	}
}


void	pose_data_scale(pose_data* accum, const float weight)
{
	assert(sizeof(pose_data) == sizeof(float) * DOF_COUNT);

	float*	result = (float*) accum;

	for (int i = 0; i < DOF_COUNT; i++)
	{
		result[i] *= weight;
	}
}


struct pose
{
	std::string	m_name;
	pose_data	m_data;
	property*	m_props;

	pose(const char* name)
		:
		m_name(name)
	{
		assert(sizeof(pose_data) == sizeof(float) * DOF_COUNT);
	}
		

	void	set(float f0, float f1, float f2, float f3, float f4, float f5, float f6, float f7)
	{
		x(0) = f0;
 		x(1) = f1;
 		x(2) = f2;
 		x(3) = f3;
 		x(4) = f4;
		x(5) = f5;
 		x(6) = f6;
		x(7) = f7;
	}

	float&	x(int i) { assert(i >= 0 && i < DOF_COUNT); return ((float*)(&m_data))[i]; }

	// *=
	// +=
};

// One pose per input bit
pose* s_rest(0);
pose* s_left(0);
pose* s_right(0);
pose* s_up(0);
pose* s_down(0);


void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}


void
ClearAll(void)
{
    ClearMouseSpring();
    
    for(int i = 0;i < objs.size();++i)
    {
        assert(objs[i]);
        simulator.remove(*objs[i]->rb);
    }

    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        simulator.remove(*forces[i]);
    }

    forces.clear();
    objs.clear();
}

void
LoadTestFile(void)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile(TestLevelFileName, 0,22, 1.0/16, 1.0/16,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

    for (int i=0; i < num_rb; ++i)
    {
        objs.push_back(new game_object("content/bitmaps/test.jpg",
                                       rb_list[i]));
    }

    for (int i=0; i < num_force; ++i)
    {
        forces.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }

}


// void	single_clank()
// {
// 	static char*	s_sounds[] =
// 	{
// 		"content/sounds/metal_a.wav",
// 		"content/sounds/metal_b.wav",
// 		"content/sounds/metal_c.wav",
// 		"content/sounds/metal_d.wav",
// 		"content/sounds/metal_e.wav",
// 		"content/sounds/metal_f.wav",
// 		"content/sounds/metal_g.wav",
// 		"content/sounds/metal_h.wav",
// 	};
	
// 	int	index = rand() % (sizeof(s_sounds)/sizeof(s_sounds[0]));
// 	zsoundPlay(s_sounds[index]);
// }


// void	soft_single_clank()
// {
// 	static char*	s_sounds[] =
// 	{
// 		"content/sounds/metal_c.wav",
// 		"content/sounds/metal_f.wav",
// 	};
	
// 	int	index = rand() % (sizeof(s_sounds)/sizeof(s_sounds[0]));
// 	zsoundPlay(s_sounds[index]);
// //	zsoundPlay(s_sounds[Globals.sound_index]);
// }


// void	multi_clank()
// {
// 	static char*	s_sounds[] =
// 	{
// 		"content/sounds/metal_a.wav",
// 		"content/sounds/metal_b.wav",
// 		"content/sounds/metal_c.wav",
// 		"content/sounds/metal_d.wav",
// 		"content/sounds/metal_e.wav",
// 		"content/sounds/metal_f.wav",
// 		"content/sounds/metal_g.wav",
// 		"content/sounds/metal_h.wav",
// 	};
	
// 	int	index = rand() % (sizeof(s_sounds)/sizeof(s_sounds[0]));
// 	zsoundPlay(s_sounds[index]);
// }


void	punch_head_sound()
{
	static char*	s_sounds[] =
	{
		"content/sounds/metal_0.wav",
	};
	int	index = rand() % (sizeof(s_sounds)/sizeof(s_sounds[0]));
	zsoundPlay(s_sounds[index]);
}


void	punch_body_sound()
{
	static char*	s_sounds[] =
	{
		"content/sounds/metal_d.wav",
	};
	int	index = rand() % (sizeof(s_sounds)/sizeof(s_sounds[0]));
	zsoundPlay(s_sounds[index]);
}


void	fell_on_head_sound()
{
	zsoundPlay("content/sounds/metal_head_bounce.wav");

	static char*	s_sounds[] =
	{
		"content/sounds/bad.0003.wav",
		"content/sounds/bad.0002.wav",
	};
	int	index = rand() % (sizeof(s_sounds)/sizeof(s_sounds[0]));
	zsoundPlay(s_sounds[index]);
}


void	fell_off_sound()
{
	zsoundPlay("content/sounds/bad.0001.wav");
}


void	punch_sound()
{
	zsoundPlay("content/sounds/metal_step3.wav");
}

void	game_over_sound()
{
	zsoundPlay("content/sounds/good.0003.wav");
}


void	game_start_sound()
{
	zsoundPlay("content/sounds/whip_1.wav");
}


void	extend_sound(int index)
{
	static char*	s_sounds[] =
	{
 		"content/sounds/metal_step6.wav",
 		"content/sounds/metal_step5.wav",
	};
	
	zsoundPlay(s_sounds[index % (sizeof(s_sounds)/sizeof(s_sounds[0]))]);
}


void	retract_sound(int index)
{
	extend_sound(index ^ 1);
}


bool	any_start_key()
{
	if (s_any_start_key) { return true; }

	for (int i = 0; i < KEY_COUNT; i++)
	{
		if (s_left_keys[i] || s_right_keys[i]) return true;
	}

	return false;
}


void	reset_players()
{
	ClearAll();
	LoadTestFile();
}


struct game_state
{
	static const int GAME_LENGTH = 75;
	static const int MSG_COUNT = 4;

	int	m_l_score;
	int	m_r_score;
	float	m_timer;
	float	m_inhibit_controls_timer;
	float	m_reset_timer;
	std::string	m_message[MSG_COUNT];
	float	m_message_timer;

	bool	m_controls_locked;

	enum phase
	{
		GAME,
		OVER,
		ATTRACT
	};
	phase	m_phase;

	game_state()
	{
		reset(ATTRACT);
		m_timer = 60.0f;
	}

	void	reset(phase new_phase)
	{
		m_l_score = 0;
		m_r_score = 0;
		m_timer = 0;
		m_inhibit_controls_timer = 0;
		m_message_timer = 0;
		m_phase = new_phase;
		m_controls_locked = false;
	}

	void	update(float dt)
	{
		m_timer += dt;
		m_message_timer -= dt;
		m_inhibit_controls_timer += dt;

		if (m_message_timer < 0)
		{
			message("");	// scroll up
		}

		if (m_phase == GAME)
		{
			if (m_timer >= float(GAME_LENGTH))
			{
				m_phase = OVER;
				m_timer = 0;

				// Show victory message.
				if (m_l_score > m_r_score)
				{
					message("         Left Wins!     ");
				}
				else if (m_r_score > m_l_score)
				{
					message("         Right Wins!     ");
				}

				game_over_sound();
			}
			else if (m_reset_timer < 0)
			{
				m_reset_timer += dt;
				if (m_reset_timer >= 0)
				{
					reset_players();
					m_inhibit_controls_timer = -0.2f;
				}
			}

			// Near the end, lock the controls together.
			float	m_switch_controls_time = GAME_LENGTH - 15.0f;
			if (m_controls_locked == false && m_timer >= m_switch_controls_time)
			{
				m_controls_locked = true;
				message("       Let's Dance!     ");
				reset_players();
				m_inhibit_controls_timer = 0.2f;
			}
		}
		else if (m_phase == OVER)
		{
			if (m_timer > 10.0f)
			{
				reset(ATTRACT);
				reset_players();
			}
			else if (m_timer > 3.0f && any_start_key())
			{
				// Start the game.
				reset(GAME);
				reset_players();
				m_inhibit_controls_timer = -1.0f;
				message("               Go!            ");
				message("");
				message("");
				message("");
				game_start_sound();
			}
		}
		else if (m_phase == ATTRACT)
		{
			if (any_start_key())
			{
				reset(GAME);
				reset_players();
				m_inhibit_controls_timer = -1.0f;
				message("               Go!            ");
				message("");
				message("");
				message("");
				game_start_sound();
			}
			else if (m_timer >= 0)
			{
				message("");
				message("      Deadly Dance Of The Robots!   ");
				message("       design by Thatcher Ulrich   ");
				message(" P1 use pad1 / [Q A E D F]");
				message(" P2 use pad2 / [Ins Del PgUp PgDn Enter]");
				m_timer -= 13.0f;

				reset_players();
			}
		}
	}

	void	display()
	{
		char	buffer[1000];

		// Show left score
		glColor3f(255/255.f, 142/255.f, 142/255.f);
//		glColor3f(1, 1, 1);
		sprintf(buffer, "%3d", m_l_score);
		DrawString(25, WindowHeight - 75.0f, buffer);

		// Show right score
		glColor3f(183/255.f, 183/255.f, 255/255.f);
//		glColor3f(1, 1, 1);
		sprintf(buffer, "%3d", m_r_score);
		DrawString(WindowWidth - 170.0f, WindowHeight - 75.0f, buffer);

		// Show message.
		{
			float	x = WindowWidth / 2.0f - 400.0f;
			float	y = WindowHeight / 2.0f + 50.0f;
			for (int i = 0; i < MSG_COUNT; i++)
			{
				glColor3f(1, 1, 1);
				DrawString(x, y, (char*) m_message[i].c_str());
				y += 50.f;
			}
		}

		if (m_phase == OVER)
		{
			glColor3f(1, 1, 1);

			// Show time
			sprintf(buffer, "time: 0");
			DrawString(float(WindowWidth / 2) - 50.0f, float(WindowHeight - 75.f), buffer);
		}
		else if (m_phase == ATTRACT)
		{
			// Show time
			sprintf(buffer, "time: %d", GAME_LENGTH);
			DrawString(float(WindowWidth / 2) - 50.0f, float(WindowHeight - 75.f), buffer);
		}
		else if (m_phase == GAME)
		{
			// Show time
			glColor3f(1, 1, 1);

			sprintf(buffer, "time: %d", int(float(GAME_LENGTH) - m_timer));
			DrawString(float(WindowWidth / 2) - 50.0f, WindowHeight - 75.f, buffer);
		}
	}


	void	message(const char* fmt, ...)
	// Show a message for moment.
	{
		char	buffer[1000];
		va_list ap;
		va_start(ap, fmt);
		_vsnprintf(buffer, 1000, fmt, ap);
		va_end(ap);
		
		// Shift up.
		for (int i = MSG_COUNT - 1; i > 0; i--)
		{
			m_message[i] = m_message[i - 1];
		}
		m_message[0] = buffer;
		m_message_timer = 3.0f;
	}
};
static game_state s_game_state;




struct robot_control
{
	float	m_leg0_extension;
	float	m_leg0_angle;
	float	m_leg1_extension;
	float	m_leg1_angle;
	float	m_punch_extension;
	float	m_arm_angle;

	float	m_foot0_friction;
	float	m_foot1_friction;

	bool	m_foot0_collide;
	bool	m_foot1_collide;

	robot_control()
		:
		m_leg0_extension(0),
		m_leg0_angle(-0.6f),
		m_leg1_extension(0),
		m_leg1_angle(0.6f),
		m_punch_extension(0),
		m_arm_angle(0.1f),
		m_foot0_friction(1.0f),
		m_foot1_friction(1.0f),
		m_foot0_collide(true),
		m_foot1_collide(true)
	{
	}
};




struct robot_controller
{
	float	m_filtered_d_leg0_extension;
	float	m_filtered_d_leg1_extension;
	float	m_prev_leg0_extension;
	float	m_prev_leg1_extension;
	float	m_prev_punch_extension;

	int	m_punched_head_frames;	// *my* head got punched
	int	m_punched_body_frames;	// *my* head got punched

	bool	m_fell_on_head;
	
	float	m_fell_on_head_timer;

	robot_controller()
		:
		m_filtered_d_leg0_extension(0),
		m_filtered_d_leg1_extension(0),
		m_prev_leg0_extension(0),
		m_prev_leg1_extension(0),
		m_prev_punch_extension(0),
		m_punched_head_frames(0),
		m_punched_body_frames(0),
		m_fell_on_head(false),
		m_fell_on_head_timer(0)
	{
	}


	void	update(robot_control* control, Sim* sim, float dt, const std::string& prefix, bool keys[KEY_COUNT])
	// Compute control outputs and other state update.
	{
		m_punched_head_frames--;
		m_punched_body_frames--;

		m_fell_on_head_timer += dt;

		// Hips.
		if (keys[K_LU] && keys[K_RU])
		{
			// Pull hips in.
			control->m_leg0_angle += 0.4f;
			control->m_leg1_angle -= 0.4f;
		}
		if (keys[K_LD] && keys[K_RD])
		{
			// Spread hips out.
			control->m_leg0_angle -= 0.4f;
			control->m_leg1_angle += 0.4f;
		}
		if (keys[K_LU] && keys[K_RD])
		{
			// Tilt.
			control->m_leg0_angle -= 0.2f;
			control->m_leg1_angle -= 0.2f;
		}
		if (keys[K_LD] && keys[K_RU])
		{
			// Tilt.
			control->m_leg0_angle += 0.2f;
			control->m_leg1_angle += 0.2f;
		}

		// Leg extension.
		if (keys[K_LU])
		{
			control->m_leg0_extension += 1.0f;
		}
		if (keys[K_LD])
		{
			control->m_leg0_extension -= 0.5f;
		}
		if (keys[K_RU])
		{
			control->m_leg1_extension += 1.0f;
		}
		if (keys[K_RD])
		{
			control->m_leg1_extension -= 0.5f;
		}

		// Arm.
		if (keys[K_PUNCH])
		{
			control->m_punch_extension = 1.0f;
		}
// 		if (keys[K_SWING])
// 		{
// 			control->m_arm_angle = -0.3f;
// 			control->m_punch_extension = -1.0f;
// 		}

		// leg sounds
		if (control->m_leg0_extension > 0.75f && m_prev_leg0_extension <= 0.75f)
		{
			if (s_game_state.m_phase == game_state::GAME)
			{
				extend_sound(0);
			}
		}
		if (control->m_leg0_extension < -0.25f && m_prev_leg0_extension >= -0.25f)
		{
			if (s_game_state.m_phase == game_state::GAME)
			{
				retract_sound(0);
			}
		}
		if (control->m_leg1_extension > 0.75f && m_prev_leg1_extension <= 0.75f)
		{
			if (s_game_state.m_phase == game_state::GAME)
			{
				extend_sound(1);
			}
		}
		if (control->m_leg1_extension < -0.25f && m_prev_leg1_extension >= -0.25f)
		{
			if (s_game_state.m_phase == game_state::GAME)
			{
				retract_sound(1);
			}
		}

		// punch sound
		if (control->m_punch_extension > 0.75f && m_prev_punch_extension <= 0.75f)
		{
			if (s_game_state.m_phase == game_state::GAME)
			{
				punch_sound();
			}
		}


		// Foot friction.
		float	d = control->m_leg0_extension - m_prev_leg0_extension;
		flowpass(&m_filtered_d_leg0_extension, Globals.m_filter_damping, fmax(0, d));
		m_prev_leg0_extension = control->m_leg0_extension;

		d = control->m_leg1_extension - m_prev_leg1_extension;
		flowpass(&m_filtered_d_leg1_extension, Globals.m_filter_damping, fmax(0, d));
		m_prev_leg1_extension = control->m_leg1_extension;

		float	balance = m_filtered_d_leg0_extension - m_filtered_d_leg1_extension;
		control->m_foot0_friction = fclamp( balance * Globals.m_balance_gain + Globals.m_balance_offset, 0.1f, 100.f);
		control->m_foot1_friction = fclamp(-balance * Globals.m_balance_gain + Globals.m_balance_offset, 0.1f, 100.f);

		m_prev_punch_extension = control->m_punch_extension;
	}
};
static robot_controller	s_left_controller;
static robot_controller	s_right_controller;



struct robot_state
{
	float	m_body_vx;
	float	m_body_vy;	//?
	float	m_body_omega;

	struct leg
	{
		float	m_angle;
		float	m_extension;
	};
	leg	m_leg[2];

	float	m_arm_angle;

	robot_state(
		float f0,
		float f1,
		float f2,
		float f3,
		float f4,
		float f5,
		float f6,
		float f7)
	{
		m_body_vx = 		f0;
		m_body_vy = 		f1;
		m_body_omega = 		f2;
		m_leg[0].m_angle = 	f3;
		m_leg[0].m_extension = 	f4;
		m_leg[1].m_angle = 	f5;
		m_leg[1].m_extension = 	f6;
		m_arm_angle = f7;
	}
};


static robot_state s_robot_rest0( 0, 0, 0, -0.6f, 0.0f, 0.6f, 0.0f, 0.0f );
static robot_state s_robot_left0( 0, 0, 0, -0.6f, 1.0f, 0.6f, 0.0f, 0.0f );
static float LEG_EXT = 2.0f;
static float PUNCH_EXT = 5.0f;

float	s_hack0(0), s_hack1(0);	//xxxxx


static void	sense_robot_feet(Sim* sim, const std::string& prefix, float* foot0, float* foot1)
// Return float depending on compression/stretch of foot to leg
// joint.  negative means joint is compressed, positive means joint is stretched.
{
	*foot0 = 0;
	*foot1 = 0;

	for (int i = 0, n = (int) sim->m_binary_forces.size(); i < n; i++)
	{
		Binary_Force*	f = sim->m_binary_forces[i];

		if (f->name() == prefix + "foot0_p")
		{
			ajb::Vec2	worldB = f->m_b.x_worldS();
			ajb::Vec2	bodyB = f->m_a.rb().world_space_to_body_space(worldB);	// pos of B's point in A's space

			*foot0 = - (bodyB.y() - f->m_a.x_bodyS().y());
		}

		if (f->name() == prefix + "foot1_p")
		{
			ajb::Vec2	worldB = f->m_b.x_worldS();
			ajb::Vec2	bodyB = f->m_a.rb().world_space_to_body_space(worldB);	// pos of B's point in A's space

			*foot1 = - (bodyB.y() - f->m_a.x_bodyS().y());
		}
	}
}



void	sample_robot_state(robot_state* r, Sim* sim, const char* prefix)
{
}


void	push_robot_state_into_sim(Sim* sim, const std::string prefix, const robot_control& control, bool negate_punch)
{
	for (int i = 0, n = (int) sim->m_binary_forces.size(); i < n; i++)
	{
		Binary_Force*	f = simulator.m_binary_forces[i];

		// Leg 0
		if (f->name() == prefix + "leg0_p0")
		{
			f->m_b.m_bs_x.set_y(flerp(control.m_leg0_extension, 2.289f, 2.289f + LEG_EXT));	// should peek these offsets startup!
		}
		if (f->name() == prefix + "leg0_p1")
		{
			f->m_b.m_bs_x.set_y(flerp(control.m_leg0_extension, 0.445f, 0.445f + LEG_EXT));
		}
		if (f->name() == prefix + "leg0_angle")
		{
			Angle_Spring_Force*	as = dynamic_cast<Angle_Spring_Force*>(f);
			if (as)
			{
				as->m_rest_len = control.m_leg0_angle;
			}
		}

		// Leg 1
		if (f->name() == prefix + "leg1_p0")
		{
			f->m_b.m_bs_x.set_y(flerp(control.m_leg1_extension, 2.289f, 2.289f + LEG_EXT));	// should peek these offsets startup!
		}
		if (f->name() == prefix + "leg1_p1")
		{
			f->m_b.m_bs_x.set_y(flerp(control.m_leg1_extension, 0.445f, 0.445f + LEG_EXT));
		}
		if (f->name() == prefix + "leg1_angle")
		{
			Angle_Spring_Force*	as = dynamic_cast<Angle_Spring_Force*>(f);
			if (as)
			{
				as->m_rest_len = control.m_leg1_angle;
			}
		}

		// Punch.
		if (f->name() == prefix + "arm_p")
		{
			float	punch = flerp(control.m_punch_extension, -1.08f, -1.08f - PUNCH_EXT);
			if (negate_punch) punch = -punch;
			f->m_b.m_bs_x.set_x(punch);	// should peek these offsets startup!
		}

		// Arm swing.
		if (f->name() == prefix + "arm_angle")
		{
			Angle_Spring_Force*	as = dynamic_cast<Angle_Spring_Force*>(f);
			if (as)
			{
				float	angle = control.m_arm_angle;
				if (negate_punch) angle = -angle;
				as->m_rest_len = angle;
			}
		}
	}

	// Feet.
	for (int i = 0, n = (int) sim->m_bodies.size(); i < n; i++)
	{
		Rigid_Body*	b = simulator.m_bodies[i];

		if (b->name() == prefix + "foot0")
		{
			b->k_friction = control.m_foot0_friction;
//			b->m_Pbody->m_collision_plane = control.m_foot0_collide ? 0xFFFFFFFF : 0;
		}
		if (b->name() == prefix + "foot1")
		{
			b->k_friction = control.m_foot1_friction;
//			b->m_Pbody->m_collision_plane = control.m_foot1_collide ? 0xFFFFFFFF : 0;
		}
	}

	// xxx hack output
	if (prefix == "l_")
	{
		s_hack0 = control.m_foot0_friction;
		s_hack1 = control.m_foot1_friction;
	}
}




GLuint
LoadTexture(char *Filename)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    verify(LoadBitmapByExtension(Filename,&result));

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}


void	set_color_for_body(Rigid_Body* rb)
{
	// Set color.
	const char*	part = rb->name().c_str() + 2;
	if (rb->name()[0] == 'l')
	{
		glColor3f(255/255.f, 142/255.f, 142/255.f);

		if (s_left_controller.m_punched_head_frames > 0
			&& strcmp(part, "head") == 0)
		{
			glColor3f(1, 0, 0);
		}
		if (s_left_controller.m_punched_body_frames > 0 && strcmp(part, "body") == 0)
		{
			glColor3f(1, 0, 0);
		}
	}
	else if (rb->name()[0] == 'r')
	{
		glColor3f(183/255.f, 183/255.f, 255/255.f);	// lighter
//		glColor3f(138/255.f, 138/255.f, 255/255.f);

		if (s_right_controller.m_punched_head_frames > 0
			&& strcmp(part, "head") == 0)
		{
			glColor3f(1, 0, 0);
		}
		if (s_right_controller.m_punched_body_frames > 0 && strcmp(part, "body") == 0)
		{
			glColor3f(1, 0, 0);
		}
	}
	else if (rb->name() == "ground")
	{
		glColor3f(218/255.0f, 205/255.0f, 50/255.0f);
	}
	else
	{
		glColor3f(0, 1, 0);
	}
}

void
glBody1v(Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glVertex2f(center.x(), center.y());
				}
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

void	refresh_property_editor()
{
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
	UpdateDisplay(GlobalsEditor);
}


bool my_collision_callback(const Body& a, const Body& b,void* rb_a, void* rb_b)
{
	Rigid_Body * rba = (Rigid_Body *)rb_a;
	Rigid_Body * rbb = (Rigid_Body *)rb_b;

	if ((rba->name()[0] == 'l' || rba->name()[0] == 'r')
		&& rba->name()[1] == '_'
		&& rba->name()[0] == rbb->name()[0])
	{
		// same robot; don't self-collide
		return false;
	}

// 	if ( a.m_collision_plane == -1 || b.m_collision_plane == -1 )
// 		return true;
	
// 	if ( a.m_collision_plane == b.m_collision_plane )
// 		return false;
	
	return true;
}



/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
    InitializeFrameRate(60.0f);

	g_collision_test_callback = my_collision_callback;

	PS2Controller.init();

    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
    
	//                                 LHA   LHK   LKA   LKK   RHA     RHK   RKA   RKK
	s_rest 		= new pose("rest");
	s_left 		= new pose("left");
	s_right 	= new pose("right");
	s_up 		= new pose("up");
	s_down 		= new pose("down");


	GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
//	glClearColor(189/255.f, 215/255.f, 236/255.f, 0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "simple_loop.sgt", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);

    LoadTestFile();
}


template<class T> void swap(T* a, T* b) { T temp = *a; *a = *b; *b = temp; }


void	score_contact(const Contact& c)
{
	Rigid_Body*	a = (Rigid_Body*) c.m_Prb_a;
	Rigid_Body*	b = (Rigid_Body*) c.m_Prb_b;

	// Canonical order.
	if (b->name() == "ground")
	{
		swap(&a, &b);
	}
	if (strcmp(a->name().c_str() + 2, "arm") == 0)
	{
		swap(&a, &b);
	}

	const int	HEAD_SHOT_SCORE = 1;
	const int	BODY_BLOW_SCORE = 2;
	if (a->name()[0] != b->name()[0])
	{
		// head vs. arm
		if (a->name() == "r_head")
		{
			if (b->name() == "l_arm")
			{
				if (s_right_controller.m_punched_head_frames <= 0)
				{
					if (s_game_state.m_phase == game_state::GAME)
					{
						s_game_state.message("     Head Shot! left +%d ", HEAD_SHOT_SCORE);
						s_game_state.m_l_score += HEAD_SHOT_SCORE;
						punch_head_sound();
					}
				}
				s_right_controller.m_punched_head_frames = 5;	// debounce hits
			}
		}
		if (a->name() == "l_head")
		{
			if (b->name() == "r_arm")
			{
				if (s_left_controller.m_punched_head_frames <= 0)
				{
					if (s_game_state.m_phase == game_state::GAME)
					{
						s_game_state.message("     Head Shot! right +%d  ", HEAD_SHOT_SCORE);
						s_game_state.m_r_score += HEAD_SHOT_SCORE;
						punch_head_sound();
					}
				}
				s_left_controller.m_punched_head_frames = 5;	// debounce hits
			}
		}

		// body vs. arm
		if (a->name() == "r_body")
		{
			if (b->name() == "l_arm")
			{
				if (s_right_controller.m_punched_body_frames <= 0
					&& s_right_controller.m_punched_head_frames <= 0)
				{
					if (s_game_state.m_phase == game_state::GAME)
					{
						s_game_state.message("     Body Blow! left +%d  ", BODY_BLOW_SCORE);
						s_game_state.m_l_score += BODY_BLOW_SCORE;
						punch_body_sound();
					}
				}
				s_right_controller.m_punched_body_frames = 5;	// debounce hits
			}
		}
		if (a->name() == "l_body")
		{
			if (b->name() == "r_arm")
			{
				if (s_left_controller.m_punched_body_frames <= 0
					&& s_left_controller.m_punched_head_frames <= 0)
				{
					if (s_game_state.m_phase == game_state::GAME)
					{
						s_game_state.message("     Body Blow! right +%d  ", BODY_BLOW_SCORE);
						s_game_state.m_r_score += BODY_BLOW_SCORE;
						punch_body_sound();
					}
				}
				s_left_controller.m_punched_body_frames = 5;	// debounce hits
			}
		}
	}

	// head vs. ground
	if (a->name() == "ground")
	{
		const int	HEAD_GROUND_SCORE = 3;
		if (b->name() == "l_head")
		{
			s_left_controller.m_fell_on_head = true;

// 			if (s_left_controller.m_punched_head_frames == 0)
// 			{
// 				fell_on_head_sound();
// 			}
			s_left_controller.m_punched_head_frames = 5;

			if (s_game_state.m_reset_timer >= 0)
			{
				if (s_game_state.m_phase == game_state::GAME)
				{
					s_game_state.message("     Fell On Head! right +%d  ", HEAD_GROUND_SCORE);
					s_game_state.m_r_score += HEAD_GROUND_SCORE;
					fell_on_head_sound();
				}
				s_game_state.m_reset_timer = -0.5f;
			}
		}
		else if (b->name() == "r_head")
		{
			s_right_controller.m_fell_on_head = true;

// 			if (s_right_controller.m_punched_head_frames == 0)
// 			{
// 				fell_on_head_sound();
// 			}
			s_right_controller.m_punched_head_frames = 5;

			if (s_game_state.m_reset_timer >= 0)
			{
				if (s_game_state.m_phase == game_state::GAME)
				{
					s_game_state.message("     Fell On Head! left +%d  ", HEAD_GROUND_SCORE);
					s_game_state.m_l_score += HEAD_GROUND_SCORE;
					fell_on_head_sound();
				}
				s_game_state.m_reset_timer = -0.5f;
			}
		}
	}
}


void	add_in_gamepad_input(bool keys_out[KEY_COUNT], bool keys_in[KEY_COUNT], Joystick& gamepad, int gp_index)
{
	// Copy to in to out.
	for (int i = 0; i < KEY_COUNT; i++) { keys_out[i] = keys_in[i]; }

	// Or in the ps2 inputs.
	if (gamepad.button_is_down(Joystick::BUTTON_RIGHT_U, gp_index)) { keys_out[K_RU] = true; }
	if (gamepad.button_is_down(Joystick::BUTTON_RIGHT_D, gp_index)) { keys_out[K_RD] = true; }
	if (gamepad.button_is_down(Joystick::BUTTON_SHOULDER_R1, gp_index)) { keys_out[K_PUNCH] = true; }
	if (gamepad.button_is_down(Joystick::BUTTON_SHOULDER_R2, gp_index)) { keys_out[K_PUNCH] = true; }
	if (gamepad.button_is_down(Joystick::BUTTON_SHOULDER_L1, gp_index)) { keys_out[K_SWING] = true; }
	if (gamepad.button_is_down(Joystick::BUTTON_SHOULDER_L2, gp_index)) { keys_out[K_SWING] = true; }
	if (gamepad.dpad_UD(gp_index) == 1) { keys_out[K_LU] = true; }
	if (gamepad.dpad_UD(gp_index) == -1) { keys_out[K_LD] = true; }

	if (gamepad.left_stick(gp_index).y() > 0.5f) { keys_out[K_LU] = true; }
	if (gamepad.left_stick(gp_index).y() < -0.5f) { keys_out[K_LD] = true; }
	if (gamepad.right_stick(gp_index).y() > 0.5f) { keys_out[K_RU] = true; }
	if (gamepad.right_stick(gp_index).y() < -0.5f) { keys_out[K_RD] = true; }

	// Flip s_any_start_key on some other helpful buttons.
	if (gamepad.button_is_down(Joystick::BUTTON_START, gp_index)) { s_any_start_key = true; }
	if (gamepad.button_is_down(Joystick::BUTTON_SELECT, gp_index)) { s_any_start_key = true; }

	for (int i = 0; i < KEY_COUNT; i++) { s_any_start_key |= keys_out[i]; }
}


void	clear_keys(bool keys[KEY_COUNT])
{
	for (int i = 0; i < KEY_COUNT; i++) { keys[i] = false; }
}


float	pseudo_noise(float t)
{
	static float	s_freq[] = {
		0.10f,
		1.20f,
		5.30f,
		0.21f,
		10.0f,
		11.0f,
	};

	float	amp = 2.0f;
	float	sum = 0.0f;
	for (int i = 0; i < sizeof(s_freq)/sizeof(s_freq[0]); i++)
	{
		sum += amp * sin(t * s_freq[i]);
		amp *= 0.8f;
	}
	return sum;
}

	
void	make_random_inputs(bool left_keys[KEY_COUNT], bool right_keys[KEY_COUNT], float dt)
{
	static float	s_timer = 0;
	s_timer += dt;

	for (int i = 0; i < KEY_COUNT; i++)
	{
		bool	key0 = pseudo_noise(s_timer + i * 1000.f) > 0.f;
		bool	key1 = pseudo_noise(s_timer + i * 1000.f + 500.f) > 0.f;

		left_keys[i] = key0;	
		right_keys[i] = key1;
	}
}


void	game_tick(float dt)
{
	s_any_start_key = false;

	PS2Controller.sample();

	// Game stuff

	// Controls.

	// collect left robot input
	bool	left_keys[KEY_COUNT];
	add_in_gamepad_input(left_keys, s_left_keys, PS2Controller, 0);

	// collect right robot input
	bool	right_keys[KEY_COUNT];
	add_in_gamepad_input(right_keys, s_right_keys, PS2Controller, 1);

	// Control combining.
	if (s_game_state.m_controls_locked)
	{
		for (int i = 0; i < KEY_COUNT; i++)
		{
			left_keys[i] = right_keys[i] = (left_keys[i] || right_keys[i]);
		}

		// No punch while locked.
		left_keys[K_PUNCH] = right_keys[K_PUNCH] = false;
	}

	// Control inhbiting.
	if (s_game_state.m_phase != game_state::GAME
		|| s_game_state.m_inhibit_controls_timer < 0.0f)
	{
		clear_keys(left_keys);
		clear_keys(right_keys);
	}

	// Attract mode -- random inputs.
	if (s_game_state.m_phase == game_state::ATTRACT)
	{
		make_random_inputs(left_keys, right_keys, dt);
	}

	// Do left actions.
	{
		robot_control	control;
		s_left_controller.update(&control, &simulator, dt, "l_", left_keys);
		push_robot_state_into_sim(&simulator, "l_", control, false);
	}

	// Do right actions.
	{
		robot_control	control;
		s_right_controller.update(&control, &simulator, dt, "r_", right_keys);
		push_robot_state_into_sim(&simulator, "r_", control, true);
	}

	// Scoring.
	s_left_controller.m_fell_on_head = false;
	s_right_controller.m_fell_on_head = false;

	// Contacts.
	for (int i = 0, n = (int) simulator.m_contacts.size(); i < n; i++)
	{
		const Contact& c = simulator.m_contacts[i];
		score_contact(c);
	}

	static const float	FELL_ON_HEAD_RESET_DELAY = 2.0f;
//		const int	HEAD_GROUND_POINTS = 5;

	if (s_left_controller.m_fell_on_head == false)
	{
		s_left_controller.m_fell_on_head_timer = 0;
	}
	else if (s_left_controller.m_fell_on_head_timer >= FELL_ON_HEAD_RESET_DELAY)
	{
//  			if (s_game_state.m_reset_timer >= 0)
//  			{
//  				s_game_state.message("     Fell On Head! right +%d  ", HEAD_GROUND_POINTS);
//  				s_game_state.m_r_score += HEAD_GROUND_POINTS;
//  				s_game_state.m_reset_timer = -0.5f;
//  			}
	}

	if (s_right_controller.m_fell_on_head == false)
	{
		s_right_controller.m_fell_on_head_timer = 0;
	}
	else if (s_right_controller.m_fell_on_head_timer >= FELL_ON_HEAD_RESET_DELAY)
	{
//  			if (s_game_state.m_reset_timer >= 0)
//  			{
//  				s_game_state.message("     Fell On Head! left +%d  ", HEAD_GROUND_POINTS);
//  				s_game_state.m_l_score += HEAD_GROUND_POINTS;
//  				s_game_state.m_reset_timer = -0.5f;
//  			}
	}

	// Falling off the platform.
	if (s_game_state.m_reset_timer >= 0)
	{
		for (int i = 0, n = (int) simulator.m_bodies.size(); i < n; i++)
		{
			static const int	FELL_OFF_SCORE = 10;

			Rigid_Body*	b = simulator.m_bodies[i];
			if (b->name() == "l_head" && b->x.y() < -100.0f)
			{
				if (s_game_state.m_phase == game_state::GAME)
				{
					s_game_state.message("     Fell off! right +%d     ", FELL_OFF_SCORE);
					s_game_state.m_r_score += FELL_OFF_SCORE;
					fell_off_sound();
				}
				reset_players();
				s_game_state.m_inhibit_controls_timer = -1.0f;
			}
			else if (b->name() == "r_head" && b->x.y() < -100.f)
			{
				if (s_game_state.m_phase == game_state::GAME)
				{
					s_game_state.message("     Fell off! left +%d  ", FELL_OFF_SCORE);
					s_game_state.m_l_score += FELL_OFF_SCORE;
					fell_off_sound();
				}
				reset_players();
				s_game_state.m_inhibit_controls_timer = -1.0f;
			}
		}
	}

	s_game_state.update(dt);
}


/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

    gravity.m_g.set_y(Globals.Gravity);
	


	/**************************** step the simulator ****************************/   

    if (spring)                             //update
        spring->m_a = Handle(world_body, world_mouse);

    int const NumberOfSimSteps = 10;
    for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
        // step as many times as you like
		float	dt = 1/(NumberOfSimSteps*GetFrameRate());
        simulator.step(dt);
        // do any game stuff you want here (like per-step collision response, etc.)
		game_tick(dt);
    }
    // now wait for the rest of the time so we're consistent on fast machines
    WaitFrameRate(GetFrameRate());

    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

//     // draw axes
//     glBegin(GL_LINES);
//     glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
//     glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
//     glEnd();

// 	// Hack output xxxxxx
// 	glColor3f(0, 0, 0);
// 	glBegin(GL_LINES);
// 	glVertex2f(-14, 0); glVertex2f(-14, s_hack0);
// 	glVertex2f(-12, 0); glVertex2f(-12, s_hack1);
// 	glEnd();
// 	//xxxxx

	// Background.
	glBegin(GL_QUADS);
	glColor3f(94/255.f, 190/255.f, 204/255.f); glVertex2f(Globals.x0, Globals.y0); glVertex2f(Globals.x1, Globals.y0);
	glColor3f(6/255.f, 99/255.f, 253/255.f); glVertex2f(Globals.x1, Globals.y1); glVertex2f(Globals.x0, Globals.y1);
	glEnd();

    // draw mouse
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	
	// iterate all the bodies, rendering them
    glColor3f(0,1,0);
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

		set_color_for_body(obj.rb);
        glBody1v(obj.rb->m_Pbody.get(), true);
    }

	// outlines
    glColor3f(0,0,0);
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

        glBody1v(obj.rb->m_Pbody.get(), false);
    }

    
//     // iterate all the forces, rendering them
//     glColor3f(1,0,0);
//     for(int i = 0;i < forces.size();++i)
//     {
//         assert(forces[i]);
//         Binary_Force *Force = forces[i];

//         glForce1v(Force);
//     }
    glForce1v(spring); 

    glDisable(GL_BLEND);

// 	glColor3f(0, 0, 1);
// 	glBegin(GL_LINES);
// 	float StickScale = 10.0f;
// 	glVertex2f(0, 0);
// 	glVertex2f(StickScale * LeftStick.x(), StickScale * LeftStick.y());
// 	glVertex2f(0, 0);
// 	glVertex2f(StickScale * RightStick.x(), StickScale * RightStick.y());
// 	glEnd();

//     glColor3f(0, 0, 0);
//     char buf[100];
//     sprintf(buf,"wm: %5.1f, %5.1f    sm: %5.1f, %5.1f",
//             world_mouse.x(),world_mouse.y(),window_mouse.x(),window_mouse.y());
//     DrawString(50,50,buf);

	s_game_state.display();

// Test controller
#if 0
	char *ButtonStrings[] =
	{
		"BUTTON_RIGHT_D",
		"BUTTON_RIGHT_U",
		"BUTTON_RIGHT_L",
		"BUTTON_RIGHT_R",
		"BUTTON_SHOULDER_R1",
		"BUTTON_SHOULDER_R2",
		"BUTTON_SHOULDER_L1",
		"BUTTON_SHOULDER_L2",
		"BUTTON_RSTICK",
		"BUTTON_LSTICK",
		"BUTTON_START",
		"BUTTON_SELECT",
	};

	int Y = 70;
	for(int ButtonIndex = 0;
		ButtonIndex < Joystick::NUM_BUTTONS;
		++ButtonIndex)
	{
		if(PS2Controller.button_is_down(ButtonIndex))
		{
			DrawString(50.f, float(Y), ButtonStrings[ButtonIndex]);
			Y += 10;
		}
	}
#endif // 0

//    DrawTimingBars();
    SwapBuffers(WindowDC);
}

/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }
        case WM_LBUTTONDOWN:
        {
			SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }
        } break;

        case WM_LBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_RBUTTON))
			{
				ReleaseCapture();
			}

            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
			SetCapture(hwnd);

			if(spring)
            {
                forces.push_back(spring);
                spring = 0;
            }
            break;
        }
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) {

// Camera rotation, useless
// 			case VK_LEFT: {
// 				CameraAngleRadians += float((5.0/180.0)*M_PI);
// 				break;
// 			}
// 			case VK_RIGHT: {
// 				CameraAngleRadians -= float((5.0/180.0)*M_PI);
// 				break;
// 			}

// 			case 'M': {
// 				// demo the audio stuff
// 				if(0x8000 & GetKeyState(VK_SHIFT)) {
// 					music.play("background");
// 				} else {
// 					zsoundPlay("content/sounds/bad.0002.wav");
// 				}
// 				break;
// 			}

//			case 'E':
			case 192:	// '`'
			{
 				if(0x8000 & GetKeyState(VK_CONTROL)) {
					launchEditor(TestLevelFileName);
					ClearAll();
					LoadTestFile();
				}
			}
			break;

// 			case 'G':
// 			{
// 				if (Globals.Gravity == 0.0f)
// 				{
// 					Globals.Gravity = -100.0f;
// 				}
// 				else
// 				{
// 					Globals.Gravity = 0.0f;
// 				}
// 				refresh_property_editor();
// 				break;
// 			}
            
			case VK_ESCAPE: {
				AppQuit = true;
				break;
			}

			case 'Q':
				s_left_keys[K_LU] = true; break;
			case 'A':
				s_left_keys[K_LD] = true; break;
			case 'W':
				s_left_keys[K_U] = true; break;
			case 'S':
				s_left_keys[K_D] = true; break;
			case 'D':
				s_left_keys[K_RD] = true; break;
			case 'E':
				s_left_keys[K_RU] = true; break;
			case 'F':
				s_left_keys[K_PUNCH] = true; break;
			case 'R':
				s_left_keys[K_SWING] = true; break;

			case VK_INSERT:
				s_right_keys[K_LU] = true; break;
			case VK_DELETE:
				s_right_keys[K_LD] = true; break;
			case VK_HOME:
				s_right_keys[K_U] = true; break;
			case VK_END:
				s_right_keys[K_D] = true; break;
			case VK_PRIOR:
				s_right_keys[K_RU] = true; break;
			case VK_NEXT:
				s_right_keys[K_RD] = true; break;
			case VK_RETURN:
				s_right_keys[K_PUNCH] = true; break;
			case 107:	// keypad +
				s_right_keys[K_SWING] = true; break;

			case VK_LEFT:
				break;
			case VK_RIGHT:
				break;
			case VK_UP:
				break;
			case VK_DOWN:
				break;

            }
            break;
		}
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
			break;

        case WM_KEYUP:
		{
			switch (wParam)
			{
			case 'Q':
 				s_left_keys[K_LU] = false; break;
			case 'A':
				s_left_keys[K_LD] = false; break;
			case 'W':
				s_left_keys[K_U] = false; break;
			case 'S':
				s_left_keys[K_D] = false; break;
			case 'D':
				s_left_keys[K_RD] = false; break;
			case 'E':
				s_left_keys[K_RU] = false; break;
			case 'F':
				s_left_keys[K_PUNCH] = false; break;
			case 'R':
				s_left_keys[K_SWING] = false; break;

			case VK_INSERT:
				s_right_keys[K_LU] = false; break;
			case VK_DELETE:
				s_right_keys[K_LD] = false; break;
			case VK_HOME:
				s_right_keys[K_U] = false; break;
			case VK_END:
				s_right_keys[K_D] = false; break;
			case VK_PRIOR:
				s_right_keys[K_RU] = false; break;
			case VK_NEXT:
				s_right_keys[K_RD] = false; break;
			case VK_RETURN:
				s_right_keys[K_PUNCH] = false; break;
			case 107:	// keypad +
				s_right_keys[K_SWING] = false; break;

			case VK_LEFT:
				s_leg_control[0] = 0.0f;	// old guy
				break;
			case VK_RIGHT:
				s_leg_control[1] = 0.0f;	// old guy
				break;
			case VK_UP:
				s_up_key = false;		// old guy
				break;
			case VK_DOWN:
				s_down_key = false;		// old guy
				break;
			}
			break;
		}

		break;

        case WM_MOUSEMOVE: {
			Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
			if(MovingCamera)
			{
                Vec2 d_world = WorldFromWindowCoords(new_window_mouse) - world_mouse;
				CameraX -= d_world.x();
				CameraY -= d_world.y();
			}
            window_mouse = new_window_mouse;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle, /* sw | */ SW_MAXIMIZE);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
